package tuffy.ground.partition;



import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.OutputStreamWriter;
import java.sql.ResultSet;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;


import org.postgresql.PGConnection;

import tuffy.db.RDB;
import tuffy.ground.Grounding;
import tuffy.mln.MarkovLogicNetwork;
import tuffy.mln.Predicate;
import tuffy.util.Config;
import tuffy.util.ExceptionMan;
import tuffy.util.UIMan;
import tuffy.util.UnionFind;

/**
 * Utilities for partitioning the MRF generated by the grounding process.
 *
 */
public class Partitioning {
	private MarkovLogicNetwork mln = null;
	private RDB db = null;

	/**
	 * Construct a partitioning worker based on the grounding
	 * result.
	 * @param g the grounding worker
	 */
	public Partitioning(Grounding g){
		mln = g.getMLN();
		db = mln.getRDB();
	}


	private double perAtomWeight(){
		return 
				16.0 // MRF collections
				+ 100.0; // GAtom
	}

	private double perClauseSharedWeight(int nlits){
		return 
				(4.0 // atom-clause index
						+ 4.0 // unsat
						+ 100.0 + nlits * 4 // GClause object
						) / nlits; 
	}


	/**
	 * Agglomeratively cluster the atoms in the MRF into partitions
	 * while heuristically minimizing the (weighted) cut size. The heuristic
	 * is to scan through all clauses in the descending-abs(weight) order.
	 * 
	 * @param ramBudgetPerPartition size bound of a partition, roughly proportional to
	 * the number of atoms and clauses in each partition
	 * 
	 * @return a partitioning scheme
	 */
	public PartitionScheme partitionAtoms(double ramBudgetPerPartition){
		UIMan.verbose(1, ">>> Partitioning atoms...");
		double maxPartRAM = ramBudgetPerPartition;
		UIMan.verbose(2, "    maxRAM = " + maxPartRAM);

		try {
			db.disableAutoCommitForNow();
			// read in set of atoms
			HashSet<Integer> qatoms = new HashSet<Integer>();
			ArrayList<Integer> atoms = new ArrayList<Integer>();
			String sql = "SELECT atomID,isQuery, isqueryevid FROM " + mln.relAtoms;
			ResultSet rs = db.query(sql);
			while(rs.next()){
				int aid = rs.getInt(1);
				atoms.add(rs.getInt(1));
				if (rs.getBoolean(2) || rs.getBoolean(3)){ 
					qatoms.add(aid);
				}
			}
			rs.close();

			// the memory usage amortized onto individual atoms
			HashMap<Integer, Double> atomWts = new HashMap<Integer, Double>();
			for(int a : atoms){
				atomWts.put(a, perAtomWeight());
			}
			sql = "SELECT lits FROM " + mln.relClauses;
			rs = db.query(sql);
			while(rs.next()){
				Integer[] lits = (Integer[])rs.getArray("lits").getArray();
				double delta = perClauseSharedWeight(lits.length);
				for(int lit : lits){
					int a = Math.abs(lit);
					double w = atomWts.get(a) + delta;
					atomWts.put(a, w);
				}
			}
			rs.close();

			// first scan, compute components and parts
			UnionFind<Integer> ufpart = new UnionFind<Integer>();
			UnionFind<Integer> ufcomp = new UnionFind<Integer>();
			ufpart.makeUnionFind(atoms, atomWts);
			ufcomp.makeUnionFind(atoms, atomWts);

			if (Config.sortWhenParitioning) {
				sql = "SELECT lits FROM " + mln.relClauses +
						" ORDER BY ABS(weight) DESC";
			} else {
				sql = "SELECT lits FROM " + mln.relClauses;
			}
			long cnt = 0;
			rs = db.query(sql);
			while(rs.next()){
				cnt ++;
				if (cnt % 1000000 == 0) {
					System.out.println(cnt);
				}
				Integer[] lits = (Integer[])rs.getArray("lits").getArray();
				if(lits.length <= 1) continue;
				double afterSize = 0;
				HashSet<Integer> roots = new HashSet<Integer>();
				for(int lit : lits){
					Integer root = ufpart.getRoot(Math.abs(lit));
					if(!roots.contains(root)){
						afterSize += ufpart.clusterWeight(root);
						roots.add(root);
					}
				}
				Integer c1 = Math.abs(lits[0]);
				for(int i=1; i<lits.length; i++){
					ufcomp.union(c1, Math.abs(lits[i]));
					//if(ProbMan.testChance(1-afterSize/maxPartSize))
					if(afterSize <= maxPartRAM){
						ufpart.union(c1, Math.abs(lits[i]));
					}
				}
			}
			rs.close();

			// allocate result data structures
			HashMap<Integer, Component> compMap = new HashMap<Integer, Component>();
			for(Integer r : ufcomp.getRoots()){
				int size = ufcomp.clusterSize(r);
				double wt = ufcomp.clusterWeight(r);
				Component comp = new Component();
				comp.numAtoms = size;
				comp.ramSize = wt;
				comp.rep = r;
				compMap.put(r, comp);
			}

			HashMap<Integer, Partition> partMap = new HashMap<Integer, Partition>();
			for(Integer r : ufpart.getRoots()){
				int size = ufpart.clusterSize(r);
				double wt = ufpart.clusterWeight(r);
				Partition part = new Partition();
				part.numAtoms = size;
				part.ramSize = wt;
				partMap.put(r, part);
				Component comp = compMap.get(ufcomp.getRoot(r));
				comp.parts.add(part);
				part.parentComponent = comp;
			}

			UIMan.verbose(2, "### Num of components = " + compMap.size());
			UIMan.verbose(2, "### Num of partitions = " + partMap.size());
			UIMan.verbose(2, ">>> Computing component/partition sizes...");

			// second scan, aggregate size info
			sql = "SELECT lits, weight FROM " + mln.relClauses;;
			rs = db.query(sql);
			cnt = 0;
			while(rs.next()){
				cnt ++;
				if (cnt % 1000000 == 0) {
					System.out.println(cnt);
				}
				Integer[] lits = (Integer[])rs.getArray("lits").getArray();
				double weight = Math.abs(rs.getDouble("weight"));
				Component comp = compMap.get(ufcomp.getRoot(Math.abs(lits[0])));
				comp.totalWeight += weight;
				comp.numClauses ++;
				comp.numPins += lits.length;
				HashSet<Partition> pset = new HashSet<Partition>();
				for(int lit : lits){
					int a = Math.abs(lit);
					Partition part = partMap.get(ufpart.getRoot(a));
					if(!pset.contains(part)){
						pset.add(part);
						part.numIncidentClauses ++;
					}
				}
				if(pset.size() > 1){
					comp.totalCutWeight += weight;
					comp.numCutClauses ++;
				}
			}
			rs.close();
			db.restoreAutoCommitState();

			// assign IDs to components and parts
			ArrayList<Component> comps = new ArrayList<Component>(compMap.values());
			Collections.sort(comps);
			int partID = 0;
			for(int i=0; i<comps.size(); i++){
				Component comp = comps.get(i);
				comp.id = i+1;
				Collections.sort(comp.parts);
				for(int j=0; j<comp.parts.size(); j++){
					Partition part = comp.parts.get(j);
					part.id = ++partID;
				}
			}
			PartitionScheme pmap = new PartitionScheme(comps);

			UIMan.verbose(2, ">>> Writing partitioning scheme into database...");

			// output clustering results
			File fbuf = new File(Config.getLoadingDir(), "partition_map" + mln.getID());
			BufferedWriter writer = new BufferedWriter(new OutputStreamWriter
					(new FileOutputStream(fbuf),"UTF8"));
			for (int qa : qatoms) {
				Component c = compMap.get(ufcomp.getRoot(qa));
				c.hasQueryAtom = true;
			}
			for(int a : atoms){
				Component c = compMap.get(ufcomp.getRoot(a));
				int compid = c.id;
				int partid = partMap.get(ufpart.getRoot(a)).id;
				writer.append(a + " , " + compid + " , " + partid + 
						"," + c.hasQueryAtom + "\n");
			}
			writer.close();
			UIMan.verbose(2, ">>> Writing " + mln.relAtomPart);
			db.dropTable(mln.relAtomPart);
			sql = "CREATE TABLE " + mln.relAtomPart + 
					"(atomID INT, compID INT, partID INT, useful BOOL)";
			db.update(sql);
			FileInputStream in = new FileInputStream(fbuf);
			PGConnection con = (PGConnection)db.getConnection();
			sql = "COPY " + mln.relAtomPart + " FROM STDIN CSV";
			con.getCopyAPI().copyIn(sql, in);
			in.close();

			// dump components info
			UIMan.verbose(2, ">>> Writing " + mln.relComponents);
			fbuf = new File(Config.getLoadingDir(), "component_list" + mln.getID());
			writer = new BufferedWriter(new OutputStreamWriter
					(new FileOutputStream(fbuf),"UTF8"));
			for(int i=0; i<comps.size(); i++){
				Component c = comps.get(i);
				writer.append(c.id + " , " + c.numAtoms + " , " + 
						c.numParts() + ", " + c.hasQueryAtom + "\n");
			}
			writer.close();
			db.dropTable(mln.relComponents);
			sql = "CREATE TABLE " + mln.relComponents +
					"(compID INT, numAtoms INT, numParts INT, hasQuery BOOL)";
			db.update(sql);
			in = new FileInputStream(fbuf);
			con = (PGConnection)db.getConnection();
			sql = "COPY " + mln.relComponents + " FROM STDIN CSV";
			con.getCopyAPI().copyIn(sql, in);
			in.close();
			
			// mark all atoms in query components
			if (Config.mark_atoms_in_useful_components) {
				UIMan.verbose(2, ">>> Marking useful tuples...");
				for (Predicate p : mln.getAllPred()) {
					if (p.isClosedWorld() && !p.hasSoftEvidence()) continue;
					UIMan.verbose(2, ">>> Marking " + p.getName());
					sql = "UPDATE " + p.getRelName() + " r SET useful=TRUE FROM " +
					mln.relAtomPart + " rap WHERE r.atomID = rap.atomID AND rap.useful";
					db.update(sql);
					UIMan.verbose(2, "    Marked " + db.getLastUpdateRowCount() + " tuples");
				}
			}

			return pmap;
		} catch (Exception e) {
			ExceptionMan.handle(e);
		}
		return null;
	}

	public PartitionScheme partitionMRF(double maxPartitionSize){
		PartitionScheme pmap = partitionAtoms(maxPartitionSize);
		partitionClauses(pmap);
		return pmap;
	}

	/**
	 * Given a partitioning scheme, partition the data accordingly.
	 * 
	 */
	public void partitionClauses(PartitionScheme pmap){
		int numParts = pmap.numParts();
		String sql;
		UIMan.verbose(1, ">>> Partitioning the MRF into " + numParts + " parts...");
		UIMan.verbose(2, ">>> Writing " + mln.relClausePart);
		db.dropTable(mln.relClausePart);
		sql = "CREATE TABLE " + mln.relClausePart +
				"(cid INT, partID INT, compID INT)";
		db.update(sql);

		sql = "INSERT INTO " + mln.relClausePart + "(cid, partID, compID) " +
				"SELECT rc.cid, ra.partID, ra.compID FROM " + 
				mln.relClauses + " rc, " + mln.relAtomPart + " ra " +
				" WHERE ABS(_random_element(rc.lits))=ra.atomID";
		db.update(sql);

	}

}
